{$mode objfpc}

unit setconv;

interface

{ 
  For a "set of x..highval", fpc_oldsetsize(ord(highval)) will return
  of the set in previous FPC versions.

  The function returns -1 if the input value is invalid.
}
function fpc_oldsetsize(highord: ptrint): ptrint;


{ 
  For a "set of lowval..highval" declared when "$packset ps" is active,
  fpc_newsetsize(ord(lowval),ord(highval),ps) will return the size of
  the set in new FPC versions. Pass 0 as packset value if no packset
  setting is used (note that while the default packset setting is 0
  in all modes at this time, the default in Delphi mode may become 1
  in future versions for increased Delphi compatibility). For ps=0,
  the new set size will be equal to the old one (but the internal
  set layout may still differ, so conversion must still be done).

  The function returns -1 if the input value is invalid.
}
function fpc_newsetsize(loword, highord: ptrint; packset: byte): ptrint;


{
  For a "set of lowval..highval" declared when "$packset ps" is active,
  "os" the set data as it was produced by an older FPC version and "ns"
  a buffer of the appropriate size as determined by fpc_newsetsize(),
  fpc_convertset(os,ns,ord(lowval),ord(highval),ps) will convert the
  set data in "os" to the new format in "ns".

  Pass 0 as packset value if no packset setting is used. (note that while
  the default packset setting is 0 in all modes at this time, the default
  in Delphi mode may become 1 in future versions for increased Delphi
  compatibility). For ps=0, the new set size will be equal to the old one
  (but the internal set layout may still differ, so conversion must still
  be done).

  The function returns true if the conversion was successful, and false
  otherwise.
}
function fpc_convertset(const oldset; out newset; loword, highord: ptrint; packset: byte): boolean;


implementation


function fpc_oldsetsize(highord: ptrint): ptrint;
begin
  if (highord <= 31) then
    fpc_oldsetsize:=4
  else if (highord <= 255) then
    fpc_oldsetsize:=32
  else
    fpc_oldsetsize:=-1;
end;


function fpc_newsetsize(loword, highord: ptrint; packset: byte): ptrint;
var
 setallocbits, setbase: ptrint;
begin
  if (packset=0) then
    fpc_newsetsize:=fpc_oldsetsize(highord)
  else
    begin
      if not(packset in [1,2,4,8]) then
        begin
          fpc_newsetsize:=-1;
          exit;
        end;
      setallocbits:=packset*8;
      setbase:=loword and not(setallocbits-1);
      fpc_newsetsize:=packset*(((highord+setallocbits)-setbase) div setallocbits);
      if (fpc_newsetsize = 3) then
        fpc_newsetsize := 4;
    end;
end;


{$r-}
{$q-}

function fpc_convertset(const oldset; out newset; loword, highord: ptrint; packset: byte): boolean;
{$ifdef FPC_BIG_ENDIAN}
type
  tset = bitpacked array[0..255] of boolean;
var
  oldbitsize, newbytesize: ptrint;
  newindex, index: ptrint;
begin
  newbytesize:=fpc_newsetsize(loword,highord,packset);
  fillchar(newset,newbytesize,0);
  oldbitsize:=fpc_oldsetsize(highord)*8;
  if (oldbitsize<0) or
     not(packset in [0,1,2,4,8]) then
    begin
      fpc_convertset:=false;
      exit;
    end;
  newindex:=loword and (ptrint(packset*8)-1);
  for index := loword to oldbitsize-1 do
    begin
      tset(newset)[newindex] :=
         tset(oldset)[(index and not 31) + (31-(index and 31))];
      inc(newindex);
    end;
  fpc_convertset:=true;
end;
{$else FPC_BIG_ENDIAN}
type
  toldset = array[0..31] of byte;
var
  oldbytesize, newbytesize: ptrint;
  oldstartindex: ptrint;
begin
  oldbytesize:=fpc_oldsetsize(highord);
  if (oldbytesize<0) or
     not(packset in [0,1,2,4,8]) then
    begin
      fpc_convertset:=false;
      exit;
    end;
  newbytesize:=fpc_newsetsize(loword,highord,packset);
  fillchar(newset,newbytesize,0);
  if (packset<>0) then
    oldstartindex:=(loword and not(ptrint(packset*8)-1)) div 8
  else
    oldstartindex:=0;
  move(toldset(oldset)[oldstartindex],newset,oldbytesize-oldstartindex);
  fpc_convertset:=true;
end;
{$endif FPC_BIG_ENDIAN}

end.

